#!/usr/bin/env python
# -*- coding: utf-8 -*-

import re
from typing import Optional


from inflex.syllable import Syllable
from inflex.term import Term
from inflex.adjective_core import (
    is_singular,
    is_plural,
    convert_to_singular,
    convert_to_plural,
)
from inflex.noun import Noun


class Adjective(Term):
    """Class for detecting and converting to adjective forms."""

    """
    Regexes to be tried before applying -er or -est.
    E.g. "pretty" is converted to "pretti" according to these regexes,
    and then "er" or "est" are appended for comparative and
    superlative respectively.
    This produces "prettier" and "prettiest".
    """
    _stem_regexes = {
        re.compile(r"[aiou]y\Z"): lambda match: match.group(),
        re.compile(r"ey\Z"): lambda match: "i",
        re.compile(r"y\Z"): lambda match: "i",
        re.compile(r"e\Z"): lambda match: "",
        # re.compile(r"([dbmnt])\Z"): lambda match: match.group(1) * 2,
    }

    _stem_double_regex = re.compile(
        r"((?:[^aeiou]|^)[aeiouy]([bcdlgkmnprstvz]))\Z")

    _possessive_regex = re.compile(
        r"\A(.*)'s?\Z", flags=re.IGNORECASE | re.MULTILINE | re.DOTALL
    )
    _possessive_inflexion = {
        # Term           0TH      1ST      2ND      3RD
        "my": {
            "singular": ["my",    "my",    "your",  "its"],
            "plural":   ["our",   "our",   "your",  "their"],
        },
        "your": {
            "singular": ["your",  "my",    "your",  "its"],
            "plural":   ["your",  "our",   "your",  "their"],
        },
        "her": {
            "singular": ["her",   "my",    "your",  "her"],
            "plural":   ["their", "our",   "your",  "their"],
        },
        "his": {
            "singular": ["his",   "my",    "your",  "his"],
            "plural":   ["their", "our",   "your",  "their"],
        },
        "its": {
            "singular": ["its",   "my",    "your",  "its"],
            "plural":   ["their", "our",   "your",  "their"],
        },
        "our": {
            "singular": ["my",    "my",    "your",  "its"],
            "plural":   ["our",   "our",   "your",  "their"],
        },
        "their": {
            "singular": ["their", "my",    "your",  "its"],
            "plural":   ["their", "our",   "your",  "their"],
        },
    }

    _comparative_conversions = {
        "good": "better",
        "well": "better",
        "best": "better",

        "bad": "worse",
        "badly": "worse",
        "ill": "worse",
        "worst": "worse",

        "far": "further",
        "furthest": "further",
        # "little": "less", # Could also be "littler", e.g. "little book" -> "even littler book"
        "many": "more",
        "much": "more",
        "most": "more",

        "least": "lesser",
    }

    _superlative_conversions = {
        "good": "best",
        "well": "best",
        "better": "best",

        "bad": "worst",
        "badly": "worst",
        "ill": "worst",
        "worse": "worst",

        "far": "furthest",
        "further": "furthest",
        # "little": "least", # We opt for littlest
        "many": "most",
        "much": "most",
        "more": "most",

        "less": "least",
        "lesser": "least",
    }

    """
    Override default methods from Term
    """

    def __init__(self, term: str):
        """Creates an Adjective instance with detection and conversion methods.

        Examples:
            >>> adj = Adjective("my")
            >>> adj.plural()
            'our'

            >>> adj = Adjective("pretty")
            >>> adj.comparative()
            'prettier'
            >>> adj.superlative()
            'prettiest'

        Note:
            Capitalisation and whitespace will be preserved between input `term` and
            generated output.

        Args:
            term (str): Input word or collocation.
        """
        super().__init__(term)

    def is_adj(self) -> bool:
        """Returns `True` only if this adjective is instantiated via `Adjective(term)`.

        Returns:
            bool: Returns `True` only if this adjective is instantiated via `Adjective(term)`.
        """
        return True

    def is_singular(self) -> bool:
        """Detect whether this adjective is in singular form.

        Returns:
            bool: True if this adjective is deemed singular.
        """
        return is_singular(self.term)

    def is_plural(self) -> bool:
        """Detect whether this adjective is in plural form.

        Returns:
            bool: True if this adjective is deemed plural.
        """
        return is_plural(self.term)

    def singular(self, person: Optional[int] = 0) -> str:
        """Returns this adjective's singular form.

        Args:
            person (Optional[int], optional): Represents the grammatical "person" (1st, 2nd, 3rd).
                This option only affects possessive adjectives. Defaults to 0.

        Returns:
            str: This adjective's singular form.
        """
        self._check_valid_person(person)
        # Is it possessive form?
        match = self._possessive_regex.match(self.term)
        if match:
            return self._reapply_whitespace(Noun(match.group(1)).singular() + "'s")

        if self.term.lower() in self._possessive_inflexion:
            return self._encase(self._possessive_inflexion[self.term.lower()]["singular"][person])

        return self._encase(convert_to_singular(self.term))

    def plural(self, person: Optional[int] = 0) -> str:
        """Returns this adjective's plural form.

        Args:
            person (Optional[int], optional): Represents the grammatical "person" (1st, 2nd, 3rd).
                This option only affects possessive adjectives. Defaults to 0.

        Returns:
            str: This adjective's plural form.
        """
        self._check_valid_person(person)
        # Is it possessive form?
        match = self._possessive_regex.match(self.term)
        if match:
            noun = Noun(match.group(1)).plural() + "'s"
            return self._reapply_whitespace(re.sub(r"s's\Z", "s'",
                                                   noun,
                                                   flags=re.MULTILINE | re.DOTALL))

        if self.term.lower() in self._possessive_inflexion:
            return self._encase(self._possessive_inflexion[self.term.lower()]["plural"][person])

        return self._encase(convert_to_plural(self.term))

    def _stem(self, term: str) -> str:
        """Stem `term` so that "-er"/"-est" can be appended for comparative/superlative forms.

        Args:
            term (str): The input word to stem.

        Returns:
            str: The stemmed version of `term`, ready for appending "-er" or "-est".
        """
        # Utility method that adjusts final consonants when they need to be doubled in inflexions.
        # Apply the first relevant transform.
        for regex in Adjective._stem_regexes:
            match = regex.search(term)
            if match:
                # Adding `term[match.end():]` is unnecessary for now,
                # but allows for more complex regexes.
                return term[:match.start()] + Adjective._stem_regexes[regex](match) +\
                    term[match.end():]

        # Get a set of known syllable counts for term
        syllable_count = Syllable.count_syllables(term)

        # Duplicate last letter if:
        if (
            # The word is certainly just one syllable, or
            1 in syllable_count
            # The word is just one syllable, or
            or (not syllable_count and Syllable.guess_if_one_syllable(term))
            # The last syllable is stressed
            or (Syllable.ends_with_stress(term))
        ) and Adjective._stem_double_regex.search(term):
            # AND the word ends in (roughly) CVC
            return term + term[-1]

        return term

    def _prepend_more_most(self, word: str) -> bool:
        """Specify whether "more" or "most" should be prepended before `word` for comp/super.

        Args:
            word (str): Input word or collocation.

        TODO: Move comparative and superlative into one method which is then called by both.
        TODO: Mine wiktionary for tests

        Returns:
            bool: True if the system believes that "more" or "most" should be prepended to
                `word` when converting it to comparative or superlative, respectively.
                Based on the deemed syllable count of `word`
        """
        syllable_counts = Syllable.count_reduced_syllables(word)
        if not syllable_counts:
            return False

        if max(syllable_counts) > 2:
            return True

        if (2 in syllable_counts and
                word and
                not word.endswith(("y", "ow", "le", "er"))
            ):
            return True

        return False

    def comparative(self, only_suffix: bool = False) -> str:
        """Returns this Adjective's comparative form.

        Examples:
            >>> adj = Adjective("pretty")
            >>> adj.comparative()
            prettier

        Note:
            "little" or "far" will fail due to having multiple options:
            * little (kid)  -> littler (kid)
            * little (food) -> less (food)
            and
            * far -> further
            * far -> farther

            Fails on e.g. "boring" or "famous"
            We convert these to "boringer" and "famouser".
            In reality they should be "more boring" and "more famous"

        TODO: Adjectives that already represent the greatest degree,
              e.g. perfect, complete, total, empty
        TODO: Adjectives that cannot be compared due to their meaning,
              e.g. pregnant, equal, ideal, British
        TODO: e.g. "quiet" is converted to "more quiet" rather than "quieter"

        TODO: Often should change the second word in "dog-friendly" type collocations.
        TODO: If abbreviation (all caps), then always prepend more

        Returns:
            str: This Adjective's comparative form.
        """
        if self.term.lower() in self._comparative_conversions.values():
            return self._encase(self.term)

        if self.term.lower() in self._comparative_conversions:
            return self._encase(self._comparative_conversions[self.term.lower()])

        # If multiple words
        pattern = re.compile(r"-| ")
        match = pattern.search(self.term)
        if match:
            term = self.term[:match.start()]
            remainder = self.term[match.end():]
            # If the form is e.g. "high and mighty",
            # then we want to convert both before and after the "and"
            if remainder.startswith("and "):
                output_format = f"{{}}{match.group()}and {{}}"
                remainder = Adjective(remainder[4:]).comparative(only_suffix)
            else:
                output_format = f"{{}}{match.group()}{{}}"

            return output_format.format(Adjective(term).comparative(only_suffix), remainder)

        if not only_suffix and (self.term.isupper() or self._prepend_more_most(self.term)):
            return f"more {self._encase(self.term)}"

        return self._encase(self._stem(self.term) + "er")

    def superlative(self, only_suffix: bool = False) -> str:
        """Returns this Adjective's superlative form.

        Examples:
            >>> adj = Adjective("pretty")
            >>> adj.superlative()
            prettiest

        Note:
            "little" or "far" will fail due to having multiple options:
            * little (kid)  -> littlest (kid)
            * little (food) -> least (food)
            and
            * far -> furthest
            * far -> farthest

            Fails on e.g. "boring" or "famous"
            We convert these to "boringest" and "famousest".
            In reality they should be "most boring" and "most famous"

        TODO: 'yest_vs_iest': [['plaguy', 'plaguyest', ['plaguiest']],
                              ['cliquy', 'cliquyest', ['cliquiest']],
                              ['heauy', 'heauyest', ['heauiest']]],
        TODO:  'est_vs_most': [['left', 'leftest', ['most left', 'leftmost']],
                              ['outer', 'outerest', ['outermost']],
                              ['inner', 'innerest', ['innermost']],
                              ['upper', 'upperest', ['uppermost']],
                              ['nether', 'netherest', ['nethermost']],
                              ['hind', 'hindest', ['hindmost']]],
        TODO: Adverbs (slowly, carefully, easily, -ly) have more/most
        TODO: More/most for title case, e.g. more American

        Returns:
            str: This Adjective's superlative form.
        """
        if self.term.lower() in self._superlative_conversions.values():
            return self._encase(self.term)

        if self.term.lower() in self._superlative_conversions:
            return self._encase(self._superlative_conversions[self.term.lower()])

        pattern = re.compile(r"-| ")
        match = pattern.search(self.term)
        if match:
            term = self.term[:match.start()]
            remainder = self.term[match.end():]
            # If the form is e.g. "high and mighty",
            # then we want to convert both before and after the "and"
            if remainder.startswith("and "):
                output_format = f"{{}}{match.group()}and {{}}"
                remainder = Adjective(remainder[4:]).superlative(only_suffix)
            else:
                output_format = f"{{}}{match.group()}{{}}"

            return output_format.format(Adjective(term).superlative(only_suffix), remainder)

        if not only_suffix and (self.term.isupper() or self._prepend_more_most(self.term)):
            return f"most {self._encase(self.term)}"

        return self._encase(self._stem(self.term) + "est")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__all__ = [
    "Term"
]

import re
from typing import Callable, Generator, List, Optional, TypeVar
import warnings

T = TypeVar("T")


def list_to_generator(input_list: List[T]) -> Generator[T, None, None]:
    """Yield element from list, repeating the very last element infinitely.

    Args:
        input_list (List[T]): List of elements.

    Yields:
        T: Element of `input_list`
    """
    while True:
        yield input_list[0]
        if len(input_list) > 1:
            input_list = input_list[1:]


class Term(object):
    """`Term` is the base class of the `Noun`, `Verb`, `Adjective` subclasses,
    and holds some default implementations of methods used across these
    subclasses.

    Method docstrings from this class are inherited to the subclasses' methods.
    """

    def _transform(func: Callable[[str], str]) -> str:
        """If `func` is called with "i" or "I", then return "I", otherwise simply call `func`

        Args:
            func (Callable[[str], str]): Function for converting the casing of an input string.

        Returns:
            str: Input string converted according to `func`'s casing rules.
        """
        return lambda word: "I" if word.lower() == "i" else func(word)

    # Supported casing formats: I, lower, Title, UPPER, Mc
    # Note that if the passed word is "i", we always output "I"
    _casing_formats = {
        "I": {
            "regex": re.compile(r"^I$"),
            "transformation": _transform(str.lower)
        },
        "lower": {
            "regex": re.compile(r"^[^A-Z]+$"),
            "transformation": _transform(str.lower)
        },
        "title": {
            "regex": re.compile(r"^[A-Z][^A-Z]+$"),
            "transformation": _transform(str.title)
        },
        "upper": {
            "regex": re.compile(r"^[A-Z]+s?$"),
            "transformation": _transform(lambda word: word[:-1].upper() + word[-1] if word.endswith(("s", "S")) else word.upper())
        },
        "Mc": {
            "regex": re.compile(r"^Mc[A-Z][^A-Z]+$"),
            "transformation": _transform(lambda word: "Mc" + word[2:].title() if word.lower().startswith("mc") else word.title())
        },
    }

    # Regex for finding a word
    _word_regex = re.compile(r"([^\r\n\t\f\v\-\' ]+)")

    # Regex for extracting whitespace before and after input
    _whitespace_regex = re.compile(r"(?P<start>^\s*).*?(?P<end>\s*$)")

    def __init__(self, term: str):
        """Creates class instance with detection and conversion methods.

        Note:
            Capitalisation and whitespace will be preserved between input `term` and
            generated output.

        Args:
            term (str): Input word or collocation.
        """
        super().__init__()
        # Whitestring strings before and after the terms
        self.start = ""
        self.end = ""
        # Default format for the separator between words
        self.spaces = None

        self.term = term.strip()

        # Extract whitespace before and after the term
        if term.startswith(" ") or term.endswith(" "):
            self.start, self.end = Term._whitespace_regex.match(term).groups()

        # If there is troublesome double whitespace, find the substrings
        # between words and normalize them
        # NOTE: Assume there are no tabs, newlines, etc. in the input terms
        if " " in self.term or "-" in self.term:
            self.spaces = re.findall(r"([\r\n\t\f\v\- ]+)", self.term)
        if "  " in self.term:
            self.term = re.sub(r"\s{2,}", " ", self.term)

    def is_noun(self) -> bool:
        """Returns `True` only if this object is instantiated via `Noun(term)`.

        Returns:
            bool: Returns `True` only if this object is instantiated via `Noun(term)`.
        """
        return False

    def is_verb(self) -> bool:
        """Returns `True` only if this object is instantiated via `Verb(term)`.

        Returns:
            bool: Returns `True` only if this object is instantiated via `Verb(term)`.
        """
        return False

    def is_adj(self) -> bool:
        """Returns `True` only if this object is instantiated via `Adjective(term)`.

        Returns:
            bool: Returns `True` only if this object is instantiated via `Adjective(term)`.
        """
        return False

    def is_singular(self) -> bool:
        """Detect whether this object is in singular form.

        Returns:
            bool: True if this object is deemed singular.
        """
        raise NotImplementedError()

    def is_plural(self) -> bool:
        """Detect whether this object is in plural form.

        Returns:
            bool: True if this object is deemed plural.
        """
        raise NotImplementedError()

    def singular(self, person: Optional[int] = 0) -> str:
        """Returns this object's singular form.

        Args:
            person (Optional[int], optional): Represents the grammatical "person" (1st, 2nd, 3rd).
                This option only affects personal and possessive pronouns, possessive adjectives,
                and verbs. Defaults to 0.

        Returns:
            str: This object's singular form.
        """
        raise NotImplementedError()

    def plural(self, person: Optional[int] = 0) -> str:
        """Returns this object's plural form.

        Args:
            person (Optional[int], optional): Represents the grammatical "person" (1st, 2nd, 3rd).
                This option only affects personal and possessive pronouns, possessive adjectives,
                and verbs. Defaults to 0.

        Returns:
            str: This object's plural form.
        """
        raise NotImplementedError()

    def classical(self) -> "Term":
        """Returns an object always inflecting in the classical/unassimilated manner.

        Examples:
            >>> Noun('cow').plural()
            'cows'
            >>> Noun('cow').unassimilated().plural()
            'kine'

        Note:
            Identical to `unassimilated()`.

        Returns:
            Term: A Term object, or a subclass thereof.
        """
        return self

    def unassimilated(self) -> "Term":
        """Returns an object always inflecting in the classical/unassimilated manner.

        Examples:
            >>> Noun('cow').plural()
            'cows'
            >>> Noun('cow').unassimilated().plural()
            'kine'

        Note:
            Identical to `classical()`.

        Returns:
            Term: A Term object, or a subclass thereof.
        """
        return self.classical()  # pragma: no cover

    def _check_valid_person(self, person: int) -> bool:
        """Return True if `person` is valid, i.e. in [0, 1, 2, 3].

        Otherwise, return False and output a warning stating that the
        `person` parameter is invalid.

        Args:
            person (int): Represents the grammatical "person" (1st, 2nd, 3rd).

        Raises:
            UserWarning: If `person` is invalid, this warning is thrown.

        Returns:
            bool: True if `person` is valid, i.e. in [0, 1, 2, 3]. False Otherwise.
        """
        if person not in [0, 1, 2, 3]:
            raise ValueError(
                "Invalid `person` parameter supplied. Valid values include 0, 1, 2, and 3.")
        return True

    def as_regex(self) -> "re.Pattern":
        """Returns a `re.Pattern` which case-insensitively matches any inflected form of the word.

        Returns:
            re.Pattern: Compiled regex object which case-insensitively matches any inflected form
                of the word.

        Examples:
            >>> Noun('cherub').as_regex()
            re.compile('cherubs|cherubim|cherub', re.IGNORECASE)
            >>> Verb('eat').as_regex()
            re.compile('eats|eating|eaten|eat|ate', re.IGNORECASE)
        """
        return re.compile("|".join(sorted(map(re.escape, {self.singular(),
                                                          self.plural()}), reverse=True)), flags=re.I)

    def __repr__(self) -> str:
        """Return `repr(self)`.

        Examples:
            >>> noun = Noun("book")
            >>> f"My noun: {noun!r}"
            "My noun: Noun('book')"
        """
        return f"{self.__class__.__name__}({self._reapply_whitespace(self.term)!r})"

    def _encase(self, target: str) -> str:
        """Apply casing from `self.term` string onto `target` string.

        TODO: Currently "show--off" -> "show----off"
        TODO: self.term as i-th and target as i-th -> I-th
            : Perhaps don't force capitalize I if followed by a hyphen.
        TODO: Let self.term as ABC and target as ABCs convert to ABCs rather than ABCS
            : Also consider that some Noun tests may be broken

        Args:
            target (str): The word or collocation on which to apply the casing
                that exists on `self.term`.

        Returns:
            str: `target`, but encased according to the patterns applied on `self.term`.
        """

        # Split off 's
        suffix = ""
        if target.endswith("'s"):
            target = target[:-2]
            suffix = "'s"

        # Special case for 'I'
        if self.term == "I" or target == "I":
            return self._reapply_whitespace(target + suffix)

        # Get list of lambda functions that correspond to the
        # casing formats for `original`.
        transformations = []
        for word in Term._word_regex.findall(self.term):
            for casing_format in Term._casing_formats.values():
                if casing_format["regex"].match(word):
                    transformations.append(casing_format["transformation"])
                    break
            else:
                # If no casing regexes matches
                transformations.append(lambda word: word)

        # If no words found in term, just return target
        if not transformations:
            return self._reapply_whitespace(target + suffix)

        # Generator that gets next transformation until there is
        # just one transformation left, after which it will
        # continuously yield that last transformation
        # Apply the transformations found in `original` to `target`
        transformations_gen = list_to_generator(transformations)
        # Phrase is target, but with the proper casing from the term applied
        phrase = Term._word_regex.sub(
            lambda match_obj: next(transformations_gen)(match_obj.group()),
            target)
        return self._reapply_whitespace(phrase + suffix)

    def _reapply_whitespace(self, phrase: str) -> str:
        """Reapply whitespace formats before, after and within a phrase.
        Based on `self.start`, `self.end` and `self.spaces` which were saved at __init__().

        Args:
            phrase (str): The word or collocation on which whitespace and hyphens are added.

        Returns:
            str: `phrase`, but with whitespace before, after and within a phrase.
        """
        if self.spaces:
            spaces_iter = iter(self.spaces)
            return self.start +\
                re.sub("-| ",
                       lambda _: next(spaces_iter),
                       phrase.strip(),
                       count=len(self.spaces)) +\
                self.end
        return self.start + phrase.strip() + self.end



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
import os
import re
from typing import Dict, List, Set


class Syllable:
    """Class with Syllable information from Carnegie Mellon University's cmudict.txt.
    This class provides several static methods which can be used to count the
    number of syllables of an input word. Another purpose is to find out whether
    the word ends with stress.
    """

    _data = {}

    @staticmethod
    def data() -> Dict[str, List[List[float]]]:
        """Lazy load syllable and stress data when requested.

        Examples:
            >>> Stress.data()
            {
                "ab": [[1], [1, 1]],
                "ababa": [[0, 1, 0], [1, 0, 0]],
                "abacha": [[1, 0, 0]],
                ...
            }

        Returns:
            Dict[str, List[List[float]]]: The dictionary key is a regular word, which maps
                to a list of syllable interpretations. Each interpretation is a list of stress
                values, where 1 means Primary stress, 0.5 means secondary stress,
                and 0 is unstressed.
        """
        if Syllable._data:
            return Syllable._data

        path = os.path.join(os.path.dirname(__file__),
                            "data/cmudict_stress.json")
        with open(path, "r", encoding="utf8") as f:
            Syllable._data = json.load(f)
        return Syllable._data

    @staticmethod
    def get_stress(word: str) -> List[List[float]]:
        """Return a list of syllable interpretations that correspond to stress values, for `word`.

        Examples:
            >>> Stress.get_stress("abdomen")
            [[0, 1, 0], [1, 0, 0]]

            This shows that the word has two interpretations, both with 3 syllables.
            In the first interpretation, the second syllable has primary stress, while
            in the second interpretation, the first syllable has primary stress.
            The other syllables are unstressed.

        Args:
            word (str): The input word.

        Returns:
            List[List[float]]: A list of syllable interpretations that correspond to stress values,
                for `word`.
        """
        try:
            return Syllable.data()[word]
        except KeyError:
            return []

    @staticmethod
    def count_syllables(word: str) -> Set[int]:
        """Return the set of valid syllable counts of `word`.

        Note:
            Some words have multiple valid syllable counts, like "abdomen".

        Args:
            word (str): The input word.

        Returns:
            Set[int]: The set of valid syllable counts of `word`.
        """
        return {len(stress) for stress in Syllable.get_stress(word)}

    @staticmethod
    def count_reduced_syllables(word: str) -> Set[int]:
        """Return the set of valid syllable counts of `word`, with "un" removed from the start.

        Args:
            word (str): The input word.

        Returns:
            Set[int]: The set of valid syllable counts of `word`, but optionally with "un" removed
            from the start of `word`.
        """
        if word.startswith("un"):
            syllables = Syllable.count_syllables(word[2:])
            if syllables:
                return syllables
        syllables = Syllable.count_syllables(word)
        if syllables:
            return syllables
        return {Syllable.guess_count_syllables(word)}

    @staticmethod
    def ends_with_stress(word: str) -> bool:
        """Returns True if all syllable interpretations of `word` end with stress.

        Note:
            Either primary or secondary stress is considered stress.

        Args:
            word (str): The input word.

        Returns:
            bool: True if all syllable interpretations of `word` end with stress.
                False otherwise.
        """
        stresses = Syllable.get_stress(word)
        return all(stress[-1] > 0 for stress in stresses if stress)

    @staticmethod
    def guess_if_one_syllable(word: str) -> bool:
        """Guess whether the word is just one Syllable.

        Args:
            word (str): The input word.

        Returns:
            bool: True if `word` is guessed to be just one syllable.
        """
        return Syllable.guess_count_syllables(word) == 1

    @staticmethod
    def guess_count_syllables(word: str) -> int:
        """Guess the number of Syllables in `word` using the number of vowel groups.

        Adapted from the pattern.en module.

        Args:
            word (str): The input word.

        Returns:
            bool: The guessed number of Syllables in `word`
        """
        # Remove trailing "e"
        while word.endswith("e"):
            word = word[:-1]
        # Convert to "CVCC"
        vc = "".join('V' if char in 'aeiouy' else 'C' for char in word)
        # Remove duplicate V's
        while "VV" in vc:
            vc = vc.replace("VV", "V")
        return vc.count("V")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import re
from typing import Optional, Tuple


from inflex.term import Term
from inflex.noun_core import (
    is_singular,
    is_plural,
    convert_to_classical_plural,
    convert_to_modern_plural,
    convert_to_singular,
)
from inflex.indefinite_core import (
    select_indefinite_article,
    prepend_indefinite_article,
)


class Noun(Term):
    """Class for detecting and converting to noun forms."""

    _noun_inflection = {
        # CASE
        #   TERM             0TH            1ST             2ND             3RD
        "nominative": {
            "i": {
                "number": "singular",
                "person": 1,
                "singular": ["I",           "I",            "you",          "it"],
                "plural":   ["we",          "we",           "you",          "they"],
            },
            "you": {
                "number": "singular",
                "person": 2,
                "singular": ["you",         "I",            "you",          "it"],
                "plural":   ["you",         "we",           "you",          "they"],
            },
            "she": {
                "number": "singular",
                "person": 3,
                "singular": ["she",         "I",            "you",          "she"],
                "plural":   ["they",        "we",           "you",          "they"],
            },
            "he": {
                "number": "singular",
                "person": 3,
                "singular": ["he",          "I",            "you",          "he"],
                "plural":   ["they",        "we",           "you",          "they"],
            },
            "it": {
                "number": "singular",
                "person": 3,
                "singular": ["it",          "I",            "you",          "it"],
                "plural":   ["they",        "we",           "you",          "they"],
            },
            "we": {
                "number": "plural",
                "person": 1,
                "singular": ["I",           "I",            "you",          "it"],
                "plural":   ["we",          "we",           "you",          "they"],
            },
            "they": {
                "number": "plural",
                "person": 3,
                "singular": ["it",          "I",            "you",          "it"],
                "plural":   ["they",        "we",           "you",          "they"],
            },
            "one": {
                "number": "singular",
                "person": 3,
                "singular": ["one",         "I",            "you",          "one"],
                "plural":   ["some",        "we",           "you",          "some"],
            },
            "this": {
                "number": "singular",
                "person": 3,
                "singular": ["this",        "this",         "this",         "this"],
                "plural":   ["these",       "these",        "these",        "these"],
            },
            "that": {
                "number": "singular",
                "person": 3,
                "singular": ["that",        "that",         "that",         "that"],
                "plural":   ["those",       "those",        "those",        "those"],
            },
            "these": {
                "number": "plural",
                "person": 3,
                "singular": ["this",        "this",         "this",         "this"],
                "plural":   ["these",       "these",        "these",        "these"],
            },
            "those": {
                "number": "plural",
                "person": 3,
                "singular": ["that",		"that",			"that",			"that"],
                "plural": ["those", "those", "those", "those"],
            },
            "who": {
                "number": "singular",
                "person": 3,
                "singular": ["who",         "who",          "who",          "who"],
                "plural":   ["who",         "who",          "who",          "who"],
            },
            "whoever": {
                "number": "singular",
                "person": 3,
                "singular": ["whoever",     "whoever",      "whoever",      "whoever"],
                "plural":   ["whoever",     "whoever",      "whoever",      "whoever"],
            },
            "whosoever": {
                "number": "singular",
                "person": 3,
                "singular": ["whosoever",   "whosoever",    "whosoever",    "whosoever"],
                "plural":   ["whosoever",   "whosoever",    "whosoever",    "whosoever"],
            },
        },
        # CASE
        #   TERM             0TH            1ST             2ND             3RD
        "objective": {
            "me": {
                "number": "singular",
                "person": 1,
                "singular": ["me",          "me",           "you",          "it"],
                "plural":   ["us",          "us",           "you",          "them"],
            },
            "you": {
                "number": "singular",
                "person": 2,
                "singular": ["you",         "me",           "you",          "it"],
                "plural":   ["you",         "us",           "you",          "them"],
            },
            "her": {
                "number": "singular",
                "person": 3,
                "singular": ["her",         "me",           "you",          "her"],
                "plural":   ["them",        "us",           "you",          "them"],
            },
            "him": {
                "number": "singular",
                "person": 3,
                "singular": ["him",         "me",           "you",          "him"],
                "plural":   ["them",        "us",           "you",          "them"],
            },
            "it": {
                "number": "singular",
                "person": 3,
                "singular": ["it",          "me",           "you",          "it"],
                "plural":   ["them",        "us",           "you",          "them"],
            },
            "one": {
                "number": "singular",
                "person": 3,
                "singular": ["one",         "me",           "you",          "one"],
                "plural":   ["some",        "us",           "you",          "some"],
            },
            "us": {
                "number": "plural",
                "person": 1,
                "singular": ["me",          "me",           "you",          "it"],
                "plural":   ["us",          "us",           "you",          "them"],
            },
            "them": {
                "number": "plural",
                "person": 3,
                "singular": ["it",          "me",           "you",          "it"],
                "plural":   ["them",        "us",           "you",          "them"],
            },
            "this": {
                "number": "singular",
                "person": 3,
                "singular": ["this",        "this",         "this",         "this"],
                "plural":   ["these",       "these",        "these",        "these"],
            },
            "that": {
                "number": "singular",
                "person": 3,
                "singular": ["that",        "that",         "that",         "that"],
                "plural":   ["those",       "those",        "those",        "those"],
            },
            "these": {
                "number": "plural",
                "person": 3,
                "singular": ["this",        "this",         "this",         "this"],
                "plural":   ["these",       "these",        "these",        "these"],
            },
            "those": {
                "number": "plural",
                "person": 3,
                "singular": ["that",        "that",         "that",         "that"],
                "plural":   ["those",       "those",        "those",        "those"],
            },
            "whom": {
                "number": "singular",
                "person": 3,
                "singular": ["whom",        "whom",         "whom",         "whom"],
                "plural":   ["whom",        "whom",         "whom",         "whom"],
            },
            "whomever": {
                "number": "singular",
                "person": 3,
                "singular": ["whomever",    "whomever",     "whomever",     "whomever"],
                "plural":   ["whomever",    "whomever",     "whomever",     "whomever"],
            },
            "whomsoever": {
                "number": "singular",
                "person": 3,
                "singular": ["whomsoever",  "whomsoever",   "whomsoever",   "whomsoever"],
                "plural":   ["whomsoever",  "whomsoever",   "whomsoever",   "whomsoever"],
            },
        },
        # CASE
        #   TERM              0TH            1ST             2ND             3RD
        "possessive": {
            "mine": {
                "number": "singular",
                "person": 1,
                "singular": ["mine",        "mine",         "yours",        "its"],
                "plural":   ["ours",        "ours",         "yours",        "theirs"],
            },
            "yours": {
                "number": "singular",
                "person": 2,
                "singular": ["yours",       "mine",         "yours",        "its"],
                "plural":   ["yours",       "ours",         "yours",        "theirs"],
            },
            "hers": {
                "number": "singular",
                "person": 3,
                "singular": ["hers",        "mine",         "yours",        "hers"],
                "plural":   ["theirs",      "ours",         "yours",        "theirs"],
            },
            "his": {
                "number": "singular",
                "person": 3,
                "singular": ["his",         "mine",         "yours",        "his"],
                "plural":   ["theirs",      "ours",         "yours",        "theirs"],
            },
            "its": {
                "number": "singular",
                "person": 3,
                "singular": ["its",         "mine",         "yours",        "its"],
                "plural":   ["theirs",      "ours",         "yours",        "theirs"],
            },
            "one's": {
                "number": "singular",
                "person": 3,
                "singular": ["one's",       "mine",         "yours",        "one's"],
                "plural":   ["theirs",      "ours",         "yours",        "theirs"],
            },
            "ours": {
                "number": "plural",
                "person": 1,
                "singular": ["mine",        "mine",         "yours",        "its"],
                "plural":   ["ours",        "ours",         "yours",        "theirs"],
            },
            "theirs": {
                "number": "plural",
                "person": 3,
                "singular": ["its",         "mine",         "yours",        "its"],
                "plural":   ["theirs",      "ours",         "yours",        "theirs"],
            },
            "whose": {
                "number": "singular",
                "person": 3,
                "singular": ["whose",       "whose",        "whose",        "whose"],
                "plural":   ["whose",       "whose",        "whose",        "whose"],
            },
            "whosever": {
                "number": "singular",
                "person": 3,
                "singular": ["whosever",    "whosever",     "whosever",     "whosever"],
                "plural":   ["whosever",    "whosever",     "whosever",     "whosever"],
            },
            "whosesoever": {
                "number": "singular",
                "person": 3,
                "singular": ["whosesoever", "whosesoever",  "whosesoever",  "whosesoever"],
                "plural":   ["whosesoever", "whosesoever",  "whosesoever",  "whosesoever"]
            },
        },
        # CASE
        #   TERM                0TH            1ST             2ND             3RD
        "reflexive": {
            "myself": {
                "number": "singular",
                "person": 1,
                "singular": ["myself",      "myself",       "yourself",     "itself"],
                "plural":   ["ourselves",   "ourselves",    "yourselves",   "themselves"],
            },
            "yourself": {
                "number": "singular",
                "person": 2,
                "singular": ["yourself",    "myself",       "yourself",     "itself"],
                "plural":   ["yourselves",  "ourselves",    "yourselves",   "themselves"],
            },
            "herself": {
                "number": "singular",
                "person": 3,
                "singular": ["herself",     "myself",       "yourself",     "herself"],
                "plural":   ["themselves",  "ourselves",    "yourselves",   "themselves"],
            },
            "himself": {
                "number": "singular",
                "person": 3,
                "singular": ["himself",     "myself",       "yourself",     "himself"],
                "plural":   ["themselves",  "ourselves",    "yourselves",   "themselves"],
            },
            "themself": {
                "number": "singular",
                "person": 3,
                "singular": ["themselves",  "myself",       "yourself",     "themselves"],
                "plural":   ["themselves",  "ourselves",    "yourselves",   "themselves"],
            },
            "itself": {
                "number": "singular",
                "person": 3,
                "singular": ["itself",      "myself",       "yourself",     "itself"],
                "plural":   ["themselves",  "ourselves",    "yourselves",   "themselves"],
            },
            "oneself": {
                "number": "singular",
                "person": 3,
                "singular": ["oneself",     "myself",       "yourself",     "oneself"],
                "plural":   ["oneselves",   "ourselves",    "yourselves",   "oneselves"],
            },
            "ourselves": {
                "number": "plural",
                "person": 1,
                "singular": ["myself",      "myself",       "yourself",     "itself"],
                "plural":   ["ourselves",   "ourselves",    "yourselves",   "themselves"],
            },
            "yourselves": {
                "number": "plural",
                "person": 2,
                "singular": ["yourself",    "myself",       "yourself",     "itself"],
                "plural":   ["yourselves",  "ourselves",    "yourselves",   "themselves"],
            },
            "themselves": {
                "number": "plural",
                "person": 3,
                "singular": ["itself",      "myself",       "yourself",     "themselves"],
                "plural":   ["themselves",  "ourselves",    "yourselves",   "themselves"],
            },
            "oneselves": {
                "number": "plural",
                "person": 3,
                "singular": ["oneself",     "myself",       "yourself",     "oneself"],
                "plural":   ["oneselves",   "ourselves",    "yourselves",   "oneselves"],
            },
        },
    }

    # Regex to detect a preposition
    _prep_regex = re.compile(
        r"""\A ( \s*(?:
              about   | above   | across  | after  | among   | around   | athwart
            | at      | before  | behind  | below  | beneath | besides?
            | between | betwixt | beyond  | but    | by      | during
            | except  | for     | from    | into   | in      | near     | off
            | of      | onto    | on      | out    | over    | since    | till
            | to      | under   | until   | unto   | upon    | within   | without | with
        )\s+)
        """, flags=re.IGNORECASE | re.VERBOSE
    )

    def __init__(self, term: str):
        """Creates a Noun instance with detection and conversion methods.

        Examples:
            >>> noun = Noun("brother")
            >>> noun.plural()
            'brothers'
            >>> noun.classical().plural()
            'brethren'

            >>> noun.is_singular()
            True

        Note:
            Capitalisation and whitespace will be preserved between input `term` and
            generated output.

        Args:
            term (str): Input word or collocation.
        """
        super().__init__(term)

        # Cached classical form of this Noun, to be lazily loaded just once.
        self._classical = None

    """
    Override default methods from Term
    """

    def is_noun(self) -> bool:
        """Returns `True` only if this noun is instantiated via `Noun(term)`.

        Returns:
            bool: Returns `True` only if this noun is instantiated via `Noun(term)`.
        """
        return True

    def is_singular(self) -> bool:
        """Detect whether this noun is in singular form.

        Returns:
            bool: True if this noun is deemed singular.
        """
        return is_singular(self.term)

    def is_plural(self) -> bool:
        """Detect whether this object is in plural form.

        Returns:
            bool: True if this object is deemed plural.
        """
        return is_plural(self.term)

    def singular(self, person: Optional[int] = 0) -> str:
        """Returns this noun's singular form.

        Args:
            person (Optional[int], optional): Represents the grammatical "person" (1st, 2nd, 3rd).
                This option only affects personal and possessive pronouns. Defaults to 0.

        Returns:
            str: This noun's singular form.
        """
        self._check_valid_person(person)
        match = Noun._prep_regex.match(self.term)

        if match:
            prep = match.group()
            term = self.term[match.end():]

            for case in ["objective", "possessive", "reflexive", "nominative"]:
                if term.lower() in Noun._noun_inflection[case]:
                    return self._encase(prep + Noun._noun_inflection[case][term.lower()]["singular"][person])

            return self._encase(prep + convert_to_singular(term))

        for case in ["nominative", "objective", "possessive", "reflexive"]:
            if self.term.lower() in Noun._noun_inflection[case]:
                return self._encase(Noun._noun_inflection[case][self.term.lower()]["singular"][person])

        return self._encase(convert_to_singular(self.term))

    def plural(self, person: Optional[int] = 0) -> str:
        """Returns this noun's plural form.

        Args:
            person (Optional[int], optional): Represents the grammatical "person" (1st, 2nd, 3rd).
                This option only affects personal and possessive pronouns. Defaults to 0.

        Returns:
            str: This noun's plural form.
        """
        self._check_valid_person(person)
        match = Noun._prep_regex.match(self.term)

        if match:
            prep = match.group()
            term = self.term[match.end():]

            for case in ["objective", "possessive", "reflexive", "nominative"]:
                if term.lower() in Noun._noun_inflection[case]:
                    return self._encase(prep + Noun._noun_inflection[case][term.lower()]["plural"][person])

            return self._encase(prep + self._convert_to_plural(term))

        for case in ["nominative", "objective", "possessive", "reflexive"]:
            if self.term.lower() in Noun._noun_inflection[case]:
                return self._encase(Noun._noun_inflection[case][self.term.lower()]["plural"][person])

        return self._encase(self._convert_to_plural(self.term))

    def _convert_to_plural(self, term) -> str:
        """The convert to plural call used by this class. Is overridden for classical nouns.

        Args:
            term (term): The input word or collocation.

        Returns:
            str: The plural form of `term`.
        """
        return convert_to_modern_plural(term)

    def classical(self) -> "ClassicalNoun":
        """Returns an object always inflecting in the classical/unassimilated manner.

        Examples:
            >>> Noun('cow').plural()
            'cows'
            >>> Noun('cow').classical().plural()
            'kine'

        Note:
            Identical to `unassimilated()`.

        Returns:
            ClassicalNoun: A Noun object that pluralises according to classical rules.
        """
        if self._classical:
            return self._classical

        # "them" is an exception, as "it -> they" and "it -> them" is ambigious
        if self.term.split()[-1].lower() in ["them", "they"]:
            self._classical = ClassicalNoun(self._encase(self.term), self)
        else:
            # TODO: self.singular() versus self.term
            # TODO: Prevent needing to encase and then re-encase
            self._classical = ClassicalNoun(self._encase(self.term), self)
        return self._classical

    def as_regex(self) -> "re.Pattern":
        """Returns a `re.Pattern` which case-insensitively matches any inflected form of the word.

        Returns:
            re.Pattern: Compiled regex object which case-insensitively matches any inflected form
                of the word.

        Examples:
            >>> Noun('cherub').as_regex()
            re.compile('cherubs|cherubim|cherub', re.IGNORECASE)
        """
        return re.compile("|".join(sorted(map(re.escape, {self.singular(),
                                                          self.plural(),
                                                          self.classical().plural()
                                                          }), reverse=True)), flags=re.I)

    """
    Methods exclusively for Noun
    """

    def indef_article(self) -> str:
        """Return the correct indefinite article ("a" or "an") for `word`.

        Args:
            word (str): Input word or collocation.

        Returns:
            str: Either "a" or "an".
        """
        return select_indefinite_article(self.term)

    def indefinite(self, count: Optional[int] = 1) -> str:
        """Prepend "a" or "an" or the number to the correct form of this Noun.

        Examples:
            >>> noun = Noun("book")
            >>> noun.indefinite(count = 1)
            'a book'
            >>> noun.indefinite(count = 3)
            '3 books'

        TODO: self.term versus self.singular()

        Args:
            count (Optional[int], optional): The number of objects on which this verb applies.
                Defaults to 1.

        Returns:
            str: The singular if `count` == 1, and the plural otherwise.
        """
        if count == 1:
            return prepend_indefinite_article(self.singular())
        return f"{count} {self.plural()}"


class ClassicalNoun(Noun):
    """Subclass of noun for detecting and converting to noun forms, with a classical plural."""

    def __init__(self, term: str, modern: Noun) -> None:
        """Creates ClassicalNoun instance with detection and conversion methods.

        Note:
            Capitalisation and whitespace will be preserved between input `term` and
            output generated via e.g. `singular`.

            The main difference to Noun is that the `plural` method is overridden.
            Another difference is that this class is generally initialized using the
            singular form of the Noun. `as_regex()` and `__repr__()` are also overridden.

        Args:
            term (str): Input word or collocation.
            modern (Noun): The Noun object from which `classical()` or `unassimilated` was
                called to create this object.
        """
        super().__init__(term)
        # Modern form of this Classical noun
        self._modern = modern

    def _convert_to_plural(self, term) -> str:
        """The convert to classical plural call used by this class.

        Args:
            term (term): The input word or collocation.

        Returns:
            str: The plural form of `term`.
        """
        return convert_to_classical_plural(term)

    def classical(self) -> "ClassicalNoun":
        """Returns an object always inflecting in the classical/unassimilated manner.

        Examples:
            >>> Noun('cow').plural()
            'cows'
            >>> Noun('cow').unassimilated().plural()
            'kine'

        Note:
            Identical to `unassimilated()`.

        Returns:
            Term: A Term object, or a subclass thereof.
        """
        return self

    def modern(self) -> "Noun":
        """Returns the non-classical modern version of this Noun.

        Examples:
            >>> noun = Noun('cow')
            >>> noun == noun.classical().modern()
            True

        Returns:
            Noun: The Noun object that will pluralize according to modern rules.
        """
        return self._modern

    def as_regex(self) -> "re.Pattern":
        """Returns a `re.Pattern` which case-insensitively matches any inflected form of the word.

        Returns:
            re.Pattern: Compiled regex object which case-insensitively matches any inflected form
                of the word.

        Examples:
            >>> Noun('brother').classical().as_regex()
            re.compile('brother|brethren', re.IGNORECASE)
        """
        return re.compile("|".join(sorted(map(re.escape, {self.singular(),
                                                          self.plural(),
                                                          }), reverse=True)), flags=re.I)

    def __repr__(self) -> str:
        """Return `repr(self)`.

        Examples:
            >>> noun = Noun("book").classical()
            >>> f"My noun: {noun!r}"
            "My noun: Noun('book').classical()"
        """
        return f"{self._modern!r}.classical()"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

##########################################
## NOTE: This module was autogenerated. ##
## Contains no user-servicable parts!!! ##
##########################################

import re

VERSION = 20210419.190714

adj_is_singular = [
    "a",
    "an",
    "her",
    "his",
    "its",
    "my",
    "that",
    "their",
    "this",
    "your"
]

adj_is_plural = [
    "our",
    "some",
    "their",
    "these",
    "those",
    "your"
]

adj_singular_of = {
    "our": "my",
    "some": "a",
    "their": "their",
    "these": "this",
    "those": "that",
    "your": "your"
}

adj_plural_of = {
    "a": "some",
    "an": "some",
    "her": "their",
    "his": "their",
    "its": "their",
    "my": "our",
    "that": "those",
    "their": "their",
    "this": "these",
    "your": "your"
}

def is_plural(word: str) -> bool:
    """Detect whether `word` is in plural form.

    Args:
        word (str): Input word or collocation.

    Returns:
        str: True if `word` is deemed plural.
    """
    return word in adj_is_plural or\
           word.lower() in adj_is_plural or\
           (word not in adj_is_singular and word.lower() not in adj_is_singular)

def is_singular(word: str) -> bool:
    """Detect whether `word` is in singular form.

    Args:
        word (str): Input word or collocation.

    Returns:
        str: True if `word` is deemed singular.
    """
    return word in adj_is_singular or\
           word.lower() in adj_is_singular or\
           (word not in adj_is_plural and word.lower() not in adj_is_plural)

def convert_to_plural(word: str) -> str:
    """Convert `word` to plural form.

    Args:
        word (str): Input word or collocation.

    Returns:
        str: The plural form of `word`.
    """
    if word in adj_plural_of:
        return adj_plural_of[word]
    if not word.islower() and word.lower() in adj_plural_of:
        return adj_plural_of[word.lower()]
    return word

def convert_to_singular(word: str) -> str:
    """Convert `word` to singular form.

    Args:
        word (str): Input word or collocation.

    Returns:
        str: The singular form of `word`.
    """
    if word in adj_singular_of:
        return adj_singular_of[word]
    if not word.islower() and word.lower() in adj_singular_of:
        return adj_singular_of[word.lower()]
    return word



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import re

# This module implements A/AN inflexion for nouns

# Special cases of A/AN
ix = re.IGNORECASE | re.VERBOSE
xms = re.VERBOSE | re.MULTILINE | re.DOTALL
ORDINAL_AN = re.compile(r"\A [aefhilmnorsx]   -?th \Z", flags=ix)
ORDINAL_A = re.compile(r"\A [bcdgjkpqtuvwyz] -?th \Z", flags=ix)
EXPLICIT_AN = re.compile(
    r"\A (?: euler | hour(?!i) | heir | honest | hono )", flags=ix)
SINGLE_AN = re.compile(r"\A [aefhilmnorsx]   \Z", flags=ix)
SINGLE_A = re.compile(r"\A [bcdgjkpqtuvwyz] \Z", flags=ix)

# This pattern matches strings of capitals (i.e. abbreviations) that
# start with a "vowel-sound" consonant followed by another consonant,
# and which are not likely to be real words

ABBREV_AN = re.compile(r"""
    \A
    (?! FJO | [HLMNS]Y.  | RY[EO] | SQU
    |   ( F[LR]? | [HL] | MN? | N | RH? | S[CHKLMNPTVW]? | X(YL)?) [AEIOU]
    )
    [FHLMNRSX][A-Z]
""", flags=xms)

# This pattern codes the beginnings of all english words begining with a
# 'Y' followed by a consonant. Any other Y-consonant prefix therefore
# implies an abbreviation...

INITIAL_Y_AN = re.compile(
    r"\A y (?: b[lor] | cl[ea] | fere | gg | p[ios] | rou | tt)", flags=ix)


def prepend_indefinite_article(word: str) -> str:
    """Prepend the indefinite article ("a" or "an") to `word`.

    Args:
        word (str): Input word or collocation.

    Returns:
        str: `word` prepended by "a" or "an".
    """
    return f"{select_indefinite_article(word)} {word}"


def select_indefinite_article(word: str) -> str:
    """Return the correct indefinite article ("a" or "an") for `word`.

    Args:
        word (str): Input word or collocation.

    Returns:
        str: Either "a" or "an".
    """
    # Handle ordinal forms: Single character followed by "-th" or "th", eg "A-th" -> "an A-th".
    if ORDINAL_A.match(word):
        return "a"
    if ORDINAL_AN.match(word):
        return "an"

    # Handle special cases: Special words (honest) or a single character, eg "a" -> "an a"
    if EXPLICIT_AN.match(word):
        return "an"
    if SINGLE_AN.match(word):
        return "an"
    if SINGLE_A.match(word):
        return "a"

    # Handle abbreviations
    if ABBREV_AN.match(word):
        return "an"
    if re.match(r"\A [aefhilmnorsx][.-]", word, flags=ix):
        return "an"
    if re.match(r"\A [a-z][.-]", word, flags=ix):
        return "a"

    # Handle consonants
    if re.match(r"\A [^aeiouy]", word, flags=ix):
        return "a"

    # Handle special vowel forms
    if re.match(r"\A e [uw]", word, flags=ix):
        return "a"
    if re.match(r"\A onc?e \b ", word, flags=ix):
        return "a"
    if re.match(r"\A uni (?: [^nmd] | mo)", word, flags=ix):
        return "a"
    if re.match(r"\A ut[th]", word, flags=ix):
        return "an"
    if re.match(r"\A u [bcfhjkqrst] [aeiou]", word, flags=ix):
        return "a"

    # Handle special capitals
    if re.match(r"\A U [NK] [AIEO]?", word, flags=re.X):
        return "a"

    # Handle vowels
    if re.match(r"\A [aeiou]", word, flags=ix):
        return "an"

    # Handle Y... (before certain consonants implies (unnaturalized) "I.." sound)
    if INITIAL_Y_AN.match(word):
        return "an"

    # Otherwise, guess "a"
    return "a"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import re
from typing import Optional, Tuple

from inflex.syllable import Syllable
from inflex.term import Term
from inflex.verb_core import (
    is_plural,
    is_singular,
    is_past,
    is_pres_part,
    is_past_part,
    convert_to_singular,
    convert_to_plural,
    convert_to_past,
    convert_to_pres_part,
    convert_to_past_part,
    plural_of,
    singular_of,
    past_of,
    pres_part_of,
    past_part_of,
)


class Verb(Term):
    """Class for detecting and converting to verb forms."""

    _prefixes = (
        'counter',
        'trans',
        'cross',
        'inter',
        'under',
        'fore',
        'back',
        'over',
        # 'post',
        'out',
        'mis',
        'for',
        'dis',
        'way',
        # 'pre',
        # 'sub',
        'un',
        'in',
        # 'be',
        'up',
        're',
        # 'co',
        # 'de',
    )

    """
    Regexes to be tried before applying -ed or -ing.
    E.g. "argue" is converted to "argu" according to these regexes,
    and then "ing" or "ed" are appended for present participle,
    and past/past participle respectively.
    This produces "arguing" and "argued".
    """
    _stem_regexes = {
        # Words ending in "fer" always duplicate their consonant,
        # e.g. "transfer" -> "transferr" (+ "ed" or "ing")
        re.compile(r"fer\Z"): lambda match: "ferr",
        # Words ending in "c" will have an extra "k" appended before
        # -ed and -ing. One exception is "arc" -> "arced".
        re.compile(r"c\Z"): lambda match: "ck",
        # Words ending in "ie" will end in "y" before appending
        # -ed or -ing.
        re.compile(r"ie\Z"): lambda match: "y",
        # Words ending with "ski" don't change,
        # and then immediately have -ed or -ing appended.
        re.compile(r"ski\Z"): lambda match: "ski",
        # Words ending with "e" prepended by anything other than an "e"
        # have that "e" stripped. e.g. "argue" -> "argu"
        re.compile(r"([^e])e\Z"): lambda match: match.group(1),
        # Words ending with "er" don't duplicate.
        re.compile(r"er\Z"): lambda match: match.group(),
        # Words ending with "en" don't duplicate,
        # unless the word is small, e.g. ken -> kenned, pen -> penned, yen -> yenned
        re.compile(r"..en\Z"): lambda match: match.group(),
        # Words ending with "on" don't duplicate.
        re.compile(r"(.[bdghklmnprstzy]on)\Z"): lambda match: match.group(1),
        # Convert duplicate "ee" into just one "e"
        re.compile(r"ee\Z"): lambda match: "e",
        # Always duplicate CVl (British English)
        re.compile(r"[^aeo][aeiuo]l\Z"): lambda match: match.group() + "l",
    }

    _stem_double_regex = re.compile(
        r"((?:[^aeiou]|^)[aeiouy]([bcdlgkmnprstvz]))\Z", re.I)

    """
    Override default methods from Term
    """

    def __init__(self, term: str):
        """Creates a Verb instance with detection and conversion methods.

        Examples:
            >>> verb = Verb("fly")
            >>> verb.singular()
            'flies'
            >>> verb.past()
            'flew'
            >>> verb.past_part()
            'flying'
            >>> verb.pres_part()
            'flown'

            >>> verb.is_plural()
            True

        Note:
            Capitalisation and whitespace will be preserved between input `term` and
            generated output.

        Args:
            term (str): Input word or collocation.
        """
        super().__init__(term)

    def is_verb(self) -> bool:
        """Returns `True` only if this verb is instantiated via `Verb(term)`.

        Returns:
            bool: Returns `True` only if this verb is instantiated via `Verb(term)`.
        """
        return True

    def is_singular(self) -> bool:
        """Detect whether this verb is in singular form.

        Returns:
            bool: True if this verb is deemed singular.
        """
        # Get first word, last section of that word (if "-" in the word)
        term, _ = self.get_subterm(self.term)

        return is_singular(term)

    def is_plural(self) -> bool:
        """Detect whether this verb is in plural form.

        Returns:
            bool: True if this verb is deemed plural.
        """
        # Get first word, last section of that word (if "-" in the word)
        term, _ = self.get_subterm(self.term)

        return is_plural(term)

    def singular(self, person: Optional[int] = 0) -> str:
        """Returns this verb's singular form.

        Args:
            person (Optional[int], optional): Represents the grammatical "person" (1st, 2nd, 3rd).
                Defaults to 0.

        Returns:
            str: This verb's singular form.
        """
        self._check_valid_person(person)

        # "To be" is special
        if self.term.lower() in ["is", "am", "are"]:
            if person == 0:
                # "are" is already singular, e.g. "they are my friend",
                # but the expected result is "is", so we opt for that.
                if self.term.lower() == "are":
                    return self._encase("is")
                return self._reapply_whitespace(self.term)
            if person == 2 or not self.is_singular():
                return self._encase("are")
            if person == 1:
                return self._encase("am")
            return self._encase("is")

        # Third person uses the "notational" singular inflection
        if person == 3 or person == 0:
            # Get first word, last section of that word (if "-" in the word)
            term, form = self.get_subterm(self.term)

            # If this term is in the list of known cases
            # TODO: This partially overlaps with `known = convert_to_singular(term)` from below
            if term.lower() in singular_of:
                return self._encase(form.format(singular_of[term.lower()]))

            # Try splitting off a prefix
            prefix, subterm = self.split_prefix(term)
            if prefix:
                known = convert_to_singular(subterm)
                if known:
                    return self._encase(form.format(prefix + known))

            # Otherwise convert the first word, last section
            known = convert_to_singular(term)
            if known:
                return self._encase(form.format(known))

            # If all else fails, return the term
            return self._reapply_whitespace(self.term)

        # First and second person always use the uninflected (i.e. "notational plural" form)
        return self.plural()

    def plural(self, person: Optional[int] = 0) -> str:
        """Returns this verb's plural form.

        Args:
            person (Optional[int], optional): Represents the grammatical "person" (1st, 2nd, 3rd).
                Defaults to 0.

        Returns:
            str: This verb's plural form.
        """
        self._check_valid_person(person)

        known = None
        # Get first word, last section of that word (if "-" in the word)
        term, form = self.get_subterm(self.term)

        # If this term is in the list of known cases
        # TODO: This partially overlaps with `known = convert_to_plural(term)` from below
        if term.lower() in plural_of:
            return self._encase(form.format(plural_of[term.lower()]))

        # Try splitting off a prefix
        prefix, subterm = self.split_prefix(term)
        known = convert_to_plural(subterm)
        if known:
            return self._encase(form.format(prefix + known))

        # Otherwise convert the first word, last section
        known = convert_to_plural(term)
        if known:
            return self._encase(form.format(known))

        # If all else fails, return the term
        return self._reapply_whitespace(self.term)

    def as_regex(self) -> "re.Pattern":
        """Returns a `re.Pattern` which case-insensitively matches any inflected form of the verb.

        Returns:
            re.Pattern: Compiled regex object which case-insensitively matches any inflected form
                of the verb.

        Examples:
            >>> Verb('eat').as_regex()
            re.compile('eats|eating|eaten|eat|ate', re.IGNORECASE)
        """
        return re.compile("|".join(sorted(map(re.escape, {self.singular(),
                                                          self.plural(),
                                                          self.past(),
                                                          self.past_part(),
                                                          self.pres_part()
                                                          }), reverse=True)), flags=re.I)

    """
    Methods exclusively for Verb
    """

    def _stem(self, term: str) -> str:
        """Stem `term` so that "-ed"/"-ing" can be appended for past and present participle forms.

        Args:
            term (str): The input word to stem.

        Returns:
            str: The stemmed version of `term`, ready for appending "-ed" or "-ing".
        """
        # Utility method that adjusts final consonants when they need to be doubled in inflexions
        # Apply the first relevant transform
        for regex in Verb._stem_regexes:
            match = regex.search(term)
            if match:
                # Adding `term[match.end():]` is unnecessary for now,
                # but allows for more complex regexes.
                return term[:match.start()] + Verb._stem_regexes[regex](match) + term[match.end():]

        # Get the last word from the term, and remove a potential prefix
        last_word = term.replace("-", " ").split()[-1]
        _, last_word = self.split_prefix(last_word)

        # Get a set of known syllable counts for last_word
        syllable_count = Syllable.count_syllables(last_word)

        # Duplicate last letter if:
        if (
            # The word is certainly just one syllable, or
            1 in syllable_count
            # The word is just one syllable, or
            or (not syllable_count and Syllable.guess_if_one_syllable(last_word))
            # The last syllable is stressed
            or (Syllable.ends_with_stress(last_word))
        ) and Verb._stem_double_regex.search(term):  # AND the word ends in (roughly) CVC
            return term + term[-1]

        return term

    def split_prefix(self, term: str) -> Tuple[str, str]:
        """Split the prefix from the term.

        Examples:
            >>> self.split_prefix("unbind")
            ("un", "bind")
            >>> self.split_prefix("mistake")
            ("mis", "take")
            >>> self.split_prefix("reappear")
            ("re", "appear")
            >>> self.split_prefix("use")
            ("", "use")

        Args:
            term (str): The input word to potentially split a prefix from.

        Returns:
            Tuple[str, str]: The first string is the prefix, the second string is the remainder.
                If the input does not have a prefix to split, then the first string is empty,
                while the second string is the full input `term`.
        """
        if term.startswith(Verb._prefixes):
            for prefix in Verb._prefixes:
                if term.startswith(prefix) and len(term[len(prefix):]) > 1:
                    return prefix, term[len(prefix):]
        return "", term

    def get_subterm(self, term: str) -> Tuple[str, str]:
        """Extract last sub-section (split by '-') of the first word.

        Examples:
            >>> self.get_subterm("aaa-bbb ccc")
            ("aaa-{} ccc", "bbb")

        Args:
            term (str): The input word to potentially split the subterm from.

        Returns:
            Tuple[str, str]: The first string is the format string, e.g. "aaa-{} ccc", while
                the second string is the last sub-section, e.g. "bbb".
        """
        form = "{}"
        if " " in term:
            split_term = term.split()
            # form = f"{{}} {' '.join(split_term[1:])}"
            form = "{}" + "".join(" " + t for t in split_term[1:])
            term = split_term[0]
        if "-" in term:
            split_term = term.split("-")
            # form = f"{'-'.join(split_term[:-1])}-" + form
            form = "".join(t + "-" for t in split_term[:-1]) + form
            term = split_term[-1]
        return term, form

    def past(self) -> str:
        """Returns this Verb's past form.

        Examples:
            >>> verb = Verb("fly")
            >>> verb.past()
            "flew"

        Returns:
            str: This Verb's past form.
        """
        known = None
        # "To be" is special
        if self.term.lower() in ["is", "am"]:
            return self._encase("was")
        if self.term.lower() == "are":
            return self._encase("were")

        # Get first word, last section of that word (if "-" in the word)
        term, form = self.get_subterm(self.term)

        # If this term is in the list of known cases
        if term.lower() in past_of:
            return self._encase(form.format(past_of[term.lower()]))

        # Try splitting off a prefix
        prefix, subterm = self.split_prefix(term)
        if prefix:
            known = convert_to_past(subterm)
            if known:
                return self._encase(form.format(prefix + known))

        # Convert the root of the term
        root, form = self.get_subterm(self.plural())
        known = convert_to_past(root)
        if known:
            return self._encase(form.format(known))

        # Otherwise use the standard pattern on the root
        known = self._stem(root) + "ed"

        return self._encase(form.format(known))

    def pres_part(self) -> str:
        """Returns this Verb's present participle form.

        Examples:
            >>> verb = Verb("fly")
            >>> verb.pres_part()
            "flying"

        Returns:
            str: This Verb's present participle form.
        """
        known = None
        # If this term is in the list of known cases
        if self.term.lower() in pres_part_of:
            return self._encase(pres_part_of[self.term.lower()])

        # Get first word, last section of that word (if "-" in the word)
        term, form = self.get_subterm(self.plural())

        # Try splitting off a prefix
        prefix, subterm = self.split_prefix(term)
        if prefix:
            known = convert_to_pres_part(subterm)
            if known:
                return self._encase(form.format(prefix + known))

        # Convert the full (sub)term
        known = convert_to_pres_part(term)

        # Otherwise use the standard pattern on the root
        if known is None:
            known = self._stem(term) + "ing"

        return self._encase(form.format(known))

    def past_part(self) -> str:
        """Returns this Verb's past participle form.

        Examples:
            >>> verb = Verb("fly")
            >>> verb.pres_part()
            "flown"

        Returns:
            str: This Verb's past participle form.
        """
        known = None
        # If this term is in the list of known cases
        if self.term.lower() in past_part_of:
            return self._encase(past_part_of[self.term.lower()])

        # Get first word, last section of that word (if "-" in the word)
        term, form = self.get_subterm(self.plural())

        # Try splitting off a prefix
        prefix, subterm = self.split_prefix(term)
        if prefix:
            known = convert_to_past_part(subterm)
            if known:
                return self._encase(form.format(prefix + known))

        # Convert the full (sub)term
        known = convert_to_past_part(term)

        # Otherwise use the standard pattern on the root
        if known is None:
            known = self._stem(term) + "ed"

        return self._encase(form.format(known))

    def is_past(self) -> bool:
        """Detect whether this Verb is in past form.

        Returns:
            bool: True if this Verb is deemed past.
        """
        # Get first word, last section of that word (if "-" in the word)
        term, _ = self.get_subterm(self.term)

        return is_past(term)

    def is_pres_part(self) -> bool:
        """Detect whether this Verb is in present participle form.

        Returns:
            bool: True if this Verb is deemed present participle.
        """
        # Get first word, last section of that word (if "-" in the word)
        term, _ = self.get_subterm(self.term)

        return is_pres_part(term)

    def is_past_part(self) -> bool:
        """Detect whether this Verb is in past participle form.

        Returns:
            bool: True if this Verb is deemed past participle.
        """
        # Get first word, last section of that word (if "-" in the word)
        term, _ = self.get_subterm(self.term)

        return is_past_part(term)

    def indefinite(self, count: Optional[int] = 1) -> str:
        """Return the singular if `count` == 1, and the plural otherwise.

        Examples:
            >>> verb = Verb("fly")
            >>> verb.indefinite(count = 1)
            'flies'
            >>> verb.indefinite(count = 3)
            'fly'

        Args:
            count (Optional[int], optional): The number of objects on which this verb applies.
                Defaults to 1.

        Returns:
            str: The singular if `count` == 1, and the plural otherwise.
        """
        if count == 1:
            return self.singular()
        return self.plural()



"""
Import the Noun, Verb and Adjective classes from the relevant files,
allowing for:
>>> from inflex import Noun
"""
__all__ = [
    "Noun",
    "Verb",
    "Adjective",
]

from inflex.noun import Noun
from inflex.verb import Verb
from inflex.adjective import Adjective
